/*global it, describe */
var assert = require("assert"),
    should = require("should"),
    Avro = require("../lib/Avro").Avro,
    protocol = require('../lib/EtpSchemas'),
    reader = new Avro.BinaryReader(),
    writer = new Avro.BinaryWriter(),
    dateTimeschema = '{"type": "record","namespace": "Energistics.Datatypes","name": "DateTime","fields":[{ "name": "time", "type": "long" },{ "name": "offset", "type": "int" }]}';
    tests = [
        /* Avro binary */                   /* Js */        /* Schema */
        [[1],                               -1,             'int'],
        [[2],                               1,              'int'],
        [[3],                               -2,             'int'],
        [[4],                               2,              'int'],
        [[0],                               0,              'int'],
        [[1],                               -1,             'long'],
        [[0x13],                            -10,            'long'],
        [[0xc7, 0x1],                       -100,           'long'],
        [[0xcf, 0x0f],                      -1000,          'long'],
        [[0x9f, 0x9c, 0x01],                -10000,         'long'],
        [[0xbf, 0x9a, 0x0c],                -100000,        'long'],
        [[0xff, 0x88, 0x7a],                -1000000,       'long'],
        [[0xff, 0xd9, 0xc4, 0x09],          -10000000,      'long'],
        [[0xff, 0x83, 0xaf, 0x5f],          -100000000,     'long'],
        [[0xff, 0xa7, 0xd6, 0xb9, 0x07],    -1000000000,    'long'],
        [[0xff, 0x8f, 0xdf, 0xc0, 0x4a],    -10000000000,   'long'],
        [[2],                               1,              'long'],
        [[0x14],                            10,             'long'],
        [[0xc8, 0x1],                       100,            'long'],
        [[0xd0, 0x0f],                      1000,           'long'],
        [[0xa0, 0x9c, 0x01],                10000,          'long'],
        [[0xc0, 0x9a, 0x0c],                100000,         'long'],
        [[0x80, 0x89, 0x7a],                1000000,        'long'],
        [[0x80, 0xda, 0xc4, 0x09],          10000000,       'long'],
        [[0x80, 0x84, 0xaf, 0x5f],          100000000,      'long'],
        [[0x80, 0xa8, 0xd6, 0xb9, 0x07],    1000000000,     'long'],
        [[0x80, 0x90, 0xdf, 0xc0, 0x4a],    10000000000,    'long'],

        // 2^53, The largest effective integer value that we can put in a double
        // without loss of precision. Negative is 1 power of 2 smaller.
        [[128, 128, 128, 128, 128, 128, 128, 32], 9007199254740992, 'long'],
        [[255, 255, 255, 255, 255, 255, 255, 15], -4503599627370496, 'long'],

        [[0],                           0,          'long'],
        [[1],                           true,       'boolean'],
        [[0],                           false,      'boolean'],
        [[6, 65, 66, 67],               "ABC",      'string'],
        [[0, 0, 0, 0, 0, 0, 0, 0 ],     0,          'double'],
        [[0, 0, 0, 0 ],                 0,          'float'],
        [[0, 0, 0, 0, 0, 0, 240, 63 ],  1,          'double'],
        [[0, 0, 128, 63 ],              1,          'float'],
        [[0, 0, 0, 0, 0, 0, 240, 191 ], -1,         'double'],
        [[0, 0, 128, 191],              -1,         'float'],
        [[0, 0, 0, 0, 0, 192, 94, 64],  123,        'double'],
        [[0, 0, 246, 66],               123,        'float'],
        [[174, 71, 225, 122, 20, 174, 243, 63 ], 1.23,   'double'],
        [[164, 112, 157, 63 ],          1.2300000190734863,       'float'],
        //[[0, 0, 0, 0, 0, 0, 248, 127 ], NaN,        'double'],
        //[[0, 0, 192, 127 ],             NaN,        'float'],
        [[138, 73, 108, 59, 130, 117, 176, 84 ], 9e+99,   'double'],
        //[[],                           null,       'null']
        [[0, 2],                         {int:1},    ['int', 'string']]
    ];

function testOneDecoding(test) {
    it('should decode [' + test[0] + '] as ' + test[1], function () {
        assert.deepEqual(test[1], reader.decode(test[2], test[0]));
    });
}

function testOneEncoding(test) {
    it('should encode ' + test[1] + ' as [' + test[0] + ']', function () {
        var result = writer.encode(test[2], test[1]),
            actual = [],
            i;
        for (i = 0; i < result.length; i++) {
            actual[i] = result[i];
        }
        assert.deepEqual(test[0], actual);
    });
}

describe('BinaryReader', function () {
    describe('#decode()', function () {
        var i;
        for (i = 0; i < tests.length; i++) {
            testOneDecoding(tests[i]);
        }
    });
});

describe('BinaryWriter', function () {
    describe('#encode()', function () {
        var i = 0;
        for (i = 0; i < tests.length; i++) {
            testOneEncoding(tests[i]);
        }
    });
});

describe('BinaryWriter', function () {
    describe('#encode()', function () {
        var size = 10000000;
        it("should handle arrays of " + size + " members", function (size) {
            var value = [], i;
            for (i = 0; i < size; i++) {
                value[i] = i;
            }
            writer.encode({type : 'array', items : 'int'}, value);
        }.bind(this, size));
    });
});

describe('SchemaCache', function () {
    describe('#find()', function () {
        it("should find schemas by protocol and message type id", function () {
            var schemaCache = new Avro.SchemaCache(protocol.types);
            var schema = schemaCache.find(0, 1);
            schema.name.should.equal("RequestSession");
        }.bind(this));
    });
});
